@@ -34,13 +34,15 @@ from ..util import (
     TRANSPORT, generate_static_pw, modhex_decode,
     modhex_encode, parse_key, parse_b32_key)
 from binascii import a2b_hex, b2a_hex
+from .. import __version__
 from ..driver_otp import YkpersError
-from ..otp import OtpController
+from ..otp import OtpController, PrepareUploadFailed, SlotConfig
 from ..scancodes import KEYBOARD_LAYOUT
 import logging
 import os
 import struct
 import click
+import webbrowser
 
 
 logger = logging.getLogger(__name__)
@@ -118,7 +120,7 @@ def otp(ctx, access_code):
       $ ykman otp chalresp --generate 2
 
     \b
-      Program a Yubico OTP credential to slot 2, using the serial as public id:
+      Program a Yubico OTP credential to slot 1, using the serial as public id:
       $ ykman otp yubiotp 1 --serial-public-id
 
     \b
@@ -242,11 +244,15 @@ def delete(ctx, slot, force):
 @click.option(
     '-G', '--generate-key', is_flag=True, required=False,
     help='Generate a random secret key. Conflicts with --key.')
+@click.option(
+    '-u', '--upload', is_flag=True, required=False,
+    help='Upload credential to YubiCloud (opens in browser). '
+    'Conflicts with --force.')
 @click_force_option
 @click.pass_context
 def yubiotp(ctx, slot, public_id, private_id, key, no_enter, force,
             serial_public_id, generate_private_id,
-            generate_key):
+            generate_key, upload):
     """
     Program a Yubico OTP credential.
 
@@ -263,6 +269,9 @@ def yubiotp(ctx, slot, public_id, private_id, key, no_enter, force,
         ctx.fail('Invalid options: --private-id conflicts with '
                  '--generate-public-id.')
 
+    if upload and force:
+        ctx.fail('Invalid options: --upload conflicts with --force.')
+
     if key and generate_key:
         ctx.fail('Invalid options: --key conflicts with --generate-key.')
 
@@ -313,13 +322,33 @@ def yubiotp(ctx, slot, public_id, private_id, key, no_enter, force,
             key = click.prompt('Enter secret key', err=True)
             key = a2b_hex(key)
 
+    if not upload and not force:
+        upload = click.confirm('Upload credential to YubiCloud?',
+                               abort=False, err=True)
+    if upload:
+        try:
+            upload_url = controller.prepare_upload_key(
+                key, public_id, private_id, serial=dev.serial,
+                user_agent='ykman/' + __version__)
+            click.echo('Upload to YubiCloud initiated successfully.')
+        except PrepareUploadFailed as e:
+            error_msg = '\n'.join(e.messages())
+            ctx.fail('Upload to YubiCloud failed.\n' + error_msg)
+
     force or click.confirm('Program an OTP credential in slot {}?'.format(slot),
                            abort=True, err=True)
+
     try:
-        controller.program_otp(slot, key, public_id, private_id, not no_enter)
+        controller.program_otp(slot, key, public_id, private_id, SlotConfig(
+            append_cr=not no_enter
+        ))
     except YkpersError as e:
         _failed_to_write_msg(ctx, e)
 
+    if upload:
+        click.echo('Opening upload form in browser: ' + upload_url)
+        webbrowser.open_new_tab(upload_url)
+
 
 @otp.command()
 @click_slot_argument
@@ -367,8 +396,9 @@ def static(
     if not force:
         _confirm_slot_overwrite(controller, slot)
     try:
-        controller.program_static(
-            slot, password, not no_enter, keyboard_layout=keyboard_layout)
+        controller.program_static(slot, password, keyboard_layout, SlotConfig(
+            append_cr=not no_enter
+        ))
     except YkpersError as e:
         _failed_to_write_msg(ctx, e)
 
@@ -519,7 +549,9 @@ def hotp(ctx, slot, key, digits, counter, no_enter, force):
         err=True)
     try:
         controller.program_hotp(
-            slot, key, counter, int(digits) == 8, not no_enter)
+            slot, key, counter, int(digits) == 8, SlotConfig(
+                append_cr=not no_enter
+            ))
     except YkpersError as e:
         _failed_to_write_msg(ctx, e)
 
@@ -542,8 +574,11 @@ def hotp(ctx, slot, key, digits, counter, no_enter, force):
     '-p', '--pacing', type=click.Choice(['0', '20', '40', '60']),
     default='0', show_default=True, help='Throttle output speed by '
     'adding a delay (in ms) between characters emitted.')
+@click.option('--use-numeric-keypad', is_flag=True, show_default=True,
+              help='Use scancodes for numeric keypad when sending digits.'
+              ' Helps with some keyboard layouts. ')
 def settings(ctx, slot, new_access_code, delete_access_code, enter, pacing,
-             force):
+             use_numeric_keypad, force):
     """
     Update the settings for a slot.
 
@@ -578,7 +613,11 @@ def settings(ctx, slot, new_access_code, delete_access_code, enter, pacing,
         pacing = int(pacing)
 
     try:
-        controller.update_settings(slot, enter=enter, pacing=pacing)
+        controller.update_settings(slot, SlotConfig(
+            append_cr=enter,
+            pacing=pacing,
+            numeric_keypad=use_numeric_keypad
+        ))
     except YkpersError as e:
         _failed_to_write_msg(ctx, e)
 
